@@ -13,7 +13,6 @@ import threading
 import time
 import uuid
 import warnings
-from collections import defaultdict
 from concurrent.futures import Future
 from pathlib import Path
 from typing import (
@@ -401,7 +400,9 @@ def _get_test_suite(
             description += f" for {repo}"
         try:
             return client.create_dataset(
-                dataset_name=test_suite_name, description=description
+                dataset_name=test_suite_name,
+                description=description,
+                metadata={"__ls_runner": "pytest"},
             )
         except ls_utils.LangSmithConflictError:
             return client.read_dataset(dataset_name=test_suite_name)
@@ -420,7 +421,8 @@ def _start_experiment(
             metadata={
                 "revision_id": ls_env.get_langchain_env_var_metadata().get(
                     "revision_id"
-                )
+                ),
+                "__ls_runner": "pytest",
             },
         )
     except ls_utils.LangSmithConflictError:
@@ -447,28 +449,28 @@ def _get_example_id(
 def _end_tests(test_suite: _LangSmithTestSuite):
     git_info = ls_env.get_git_info() or {}
     test_suite.shutdown()
-    dataset_version = test_suite.get_version()
+    dataset_version = test_suite.get_dataset_version()
     dataset_id = test_suite._dataset.id
     test_suite.client.update_project(
         test_suite.experiment_id,
-        end_time=datetime.datetime.now(datetime.timezone.utc),
         metadata={
             **git_info,
             "dataset_version": dataset_version,
             "revision_id": ls_env.get_langchain_env_var_metadata().get("revision_id"),
+            "__ls_runner": "pytest",
         },
     )
     if dataset_version and git_info["commit"] is not None:
         test_suite.client.update_dataset_tag(
             dataset_id=dataset_id,
             as_of=dataset_version,
-            tag=f'git:commit:{git_info["commit"]}',
+            tag=f"git:commit:{git_info['commit']}",
         )
     if dataset_version and git_info["branch"] is not None:
         test_suite.client.update_dataset_tag(
             dataset_id=dataset_id,
             as_of=dataset_version,
-            tag=f'git:branch:{git_info["branch"]}',
+            tag=f"git:branch:{git_info['branch']}",
         )
 
 
@@ -495,10 +497,8 @@ class _LangSmithTestSuite:
         self.client = client or rt.get_cached_client()
         self._experiment = experiment
         self._dataset = dataset
-        self._version: Optional[datetime.datetime] = None
-        self._executor = ls_utils.ContextThreadPoolExecutor(max_workers=1)
-        self._example_futures: dict[ID_TYPE, list[Future]] = defaultdict(list)
-        self._example_modified_at: dict[ID_TYPE, datetime.datetime] = {}
+        self._dataset_version: Optional[datetime.datetime] = dataset.modified_at
+        self._executor = ls_utils.ContextThreadPoolExecutor()
         atexit.register(_end_tests, self)
 
     @property
@@ -535,20 +535,8 @@ class _LangSmithTestSuite:
     def name(self):
         return self._experiment.name
 
-    def update_version(self, version: datetime.datetime, example_id: uuid.UUID) -> None:
-        with self._lock:
-            if self._version is None or version > self._version:
-                self._version = version
-            self._example_modified_at[example_id] = version
-
-    def get_version(
-        self, example_id: Optional[uuid.UUID] = None
-    ) -> Optional[datetime.datetime]:
-        with self._lock:
-            if not example_id:
-                return self._version
-            else:
-                return self._example_modified_at[example_id]
+    def get_dataset_version(self):
+        return self._dataset_version
 
     def submit_result(
         self,
@@ -572,7 +560,9 @@ class _LangSmithTestSuite:
         self._executor.submit(self._submit_result, run_id, score)
 
     def _submit_result(self, run_id: uuid.UUID, score: Optional[int]) -> None:
-        self.client.create_feedback(run_id, key="pass", score=score)
+        # trace_id will always be run_id here because the feedback is on the root
+        # test run
+        self.client.create_feedback(run_id, key="pass", score=score, trace_id=run_id)
 
     def sync_example(
         self,
@@ -584,27 +574,12 @@ class _LangSmithTestSuite:
         pytest_plugin=None,
         pytest_nodeid=None,
     ) -> None:
+        inputs = inputs or {}
         if pytest_plugin and pytest_nodeid:
             update = {"inputs": inputs, "reference_outputs": outputs}
             update = {k: v for k, v in update.items() if v is not None}
             pytest_plugin.update_process_status(pytest_nodeid, update)
-        future = self._executor.submit(
-            self._sync_example,
-            example_id,
-            inputs,
-            outputs,
-            metadata.copy() if metadata else metadata,
-        )
-        with self._lock:
-            self._example_futures[example_id].append(future)
-
-    def _sync_example(
-        self,
-        example_id: uuid.UUID,
-        inputs: Optional[dict],
-        outputs: Optional[dict],
-        metadata: Optional[dict],
-    ) -> None:
+        metadata = metadata.copy() if metadata else metadata
         inputs = _serde_example_values(inputs)
         outputs = _serde_example_values(outputs)
         try:
@@ -618,26 +593,29 @@ class _LangSmithTestSuite:
                 metadata=metadata,
                 created_at=self._experiment.start_time,
             )
-            modified_at = example.modified_at
         else:
             if (
-                (inputs is not None and inputs != example.inputs)
+                (inputs != example.inputs)
                 or (outputs is not None and outputs != example.outputs)
                 or (metadata is not None and metadata != example.metadata)
                 or str(example.dataset_id) != str(self.id)
             ):
-                response = self.client.update_example(
+                self.client.update_example(
                     example_id=example.id,
                     inputs=inputs,
                     outputs=outputs,
                     metadata=metadata,
                     dataset_id=self.id,
                 )
-                modified_at = datetime.datetime.fromisoformat(response["modified_at"])
-            else:
-                modified_at = example.modified_at
-        if modified_at:
-            self.update_version(modified_at, example_id=example_id)
+                example = self.client.read_example(example_id=example.id)
+        if self._dataset_version is None:
+            self._dataset_version = example.modified_at
+        elif (
+            example.modified_at
+            and self._dataset_version
+            and example.modified_at > self._dataset_version
+        ):
+            self._dataset_version = example.modified_at
 
     def _submit_feedback(
         self,
@@ -659,22 +637,19 @@ class _LangSmithTestSuite:
             )
 
     def _create_feedback(self, run_id: ID_TYPE, feedback: dict, **kwargs: Any) -> None:
-        self.client.create_feedback(run_id, **feedback, **kwargs)
+        # trace_id will always be run_id here because the feedback is on the root
+        # test run
+        self.client.create_feedback(run_id, **feedback, **kwargs, trace_id=run_id)
 
     def shutdown(self):
         self._executor.shutdown()
 
-    def wait_example_updates(self, example_id: ID_TYPE):
-        """Wait for all example updates to complete."""
-        while self._example_futures[example_id]:
-            self._example_futures[example_id].pop().result()
-
     def end_run(
         self,
         run_tree,
         example_id,
         outputs,
-        end_time,
+        reference_outputs,
         pytest_plugin=None,
         pytest_nodeid=None,
     ) -> Future:
@@ -683,23 +658,25 @@ class _LangSmithTestSuite:
             run_tree=run_tree,
             example_id=example_id,
             outputs=outputs,
-            end_time=end_time,
+            reference_outputs=reference_outputs,
             pytest_plugin=pytest_plugin,
             pytest_nodeid=pytest_nodeid,
         )
 
     def _end_run(
-        self, run_tree, example_id, outputs, end_time, pytest_plugin, pytest_nodeid
+        self,
+        run_tree,
+        example_id,
+        outputs,
+        reference_outputs,
+        pytest_plugin,
+        pytest_nodeid,
     ) -> None:
         # TODO: remove this hack so that run durations are correct
         # Ensure example is fully updated
-        self.wait_example_updates(example_id)
-        # Ensure that run end time is after example modified at.
-        example_modified_at = self.get_version(example_id=example_id)
-        end_time = max(example_modified_at, end_time)
-        run_tree.end(outputs=outputs, end_time=end_time)
+        self.sync_example(example_id, inputs=run_tree.inputs, outputs=reference_outputs)
+        run_tree.end(outputs=outputs)
         run_tree.patch()
-        pytest_plugin.update_process_status(pytest_nodeid, {"logged": True})
 
 
 class _TestCase:
@@ -710,17 +687,26 @@ class _TestCase:
         run_id: uuid.UUID,
         pytest_plugin: Any = None,
         pytest_nodeid: Any = None,
+        inputs: Optional[dict] = None,
+        reference_outputs: Optional[dict] = None,
     ) -> None:
         self.test_suite = test_suite
         self.example_id = example_id
         self.run_id = run_id
         self.pytest_plugin = pytest_plugin
         self.pytest_nodeid = pytest_nodeid
+        self._logged_reference_outputs: Optional[dict] = None
+        self.inputs = inputs
+        self.reference_outputs = reference_outputs
 
         if pytest_plugin and pytest_nodeid:
             pytest_plugin.add_process_to_test_suite(
                 test_suite._dataset.name, pytest_nodeid
             )
+            if inputs:
+                self.log_inputs(inputs)
+            if reference_outputs:
+                self.log_reference_outputs(reference_outputs)
 
     def sync_example(
         self, *, inputs: Optional[dict] = None, outputs: Optional[dict] = None
@@ -745,12 +731,25 @@ class _TestCase:
             },
         )
 
+    def log_inputs(self, inputs: dict) -> None:
+        if self.pytest_plugin and self.pytest_nodeid:
+            self.pytest_plugin.update_process_status(
+                self.pytest_nodeid, {"inputs": inputs}
+            )
+
     def log_outputs(self, outputs: dict) -> None:
         if self.pytest_plugin and self.pytest_nodeid:
             self.pytest_plugin.update_process_status(
                 self.pytest_nodeid, {"outputs": outputs}
             )
 
+    def log_reference_outputs(self, reference_outputs: dict) -> None:
+        self._logged_reference_outputs = reference_outputs
+        if self.pytest_plugin and self.pytest_nodeid:
+            self.pytest_plugin.update_process_status(
+                self.pytest_nodeid, {"reference_outputs": reference_outputs}
+            )
+
     def submit_test_result(
         self,
         error: Optional[str] = None,
@@ -776,15 +775,14 @@ class _TestCase:
                 self.pytest_nodeid, {"end_time": time.time()}
             )
 
-    def end_run(self, run_tree, outputs: Any) -> Future:
+    def end_run(self, run_tree, outputs: Any) -> None:
         if not (outputs is None or isinstance(outputs, dict)):
             outputs = {"output": outputs}
-        end_time = datetime.datetime.now(datetime.timezone.utc)
-        return self.test_suite.end_run(
+        self.test_suite.end_run(
             run_tree,
             self.example_id,
             outputs,
-            end_time=end_time,
+            reference_outputs=self._logged_reference_outputs,
             pytest_plugin=self.pytest_plugin,
             pytest_nodeid=self.pytest_nodeid,
         )
@@ -836,12 +834,6 @@ def _create_test_case(
     )
     example_id, example_name = _get_example_id(func, inputs, test_suite.id)
     example_id = langtest_extra["id"] or example_id
-    test_suite.sync_example(
-        example_id,
-        inputs=inputs,
-        outputs=outputs,
-        metadata={"signature": _get_test_repr(func, signature), "name": example_name},
-    )
     pytest_plugin = (
         pytest_request.config.pluginmanager.get_plugin("langsmith_output_plugin")
         if pytest_request
@@ -854,13 +846,16 @@ def _create_test_case(
             + "/compare?selectedSessions="
             + str(test_suite.experiment_id)
         )
-    return _TestCase(
+    test_case = _TestCase(
         test_suite,
         example_id,
         run_id=uuid.uuid4(),
+        inputs=inputs,
+        reference_outputs=outputs,
         pytest_plugin=pytest_plugin,
         pytest_nodeid=pytest_nodeid,
     )
+    return test_case
 
 
 def _run_test(
@@ -878,8 +873,6 @@ def _run_test(
         langtest_extra=langtest_extra,
     )
     _TEST_CASE.set(test_case)
-    func_sig = inspect.signature(func)
-    func_inputs = rh._get_inputs_safe(func_sig, *test_args, **test_kwargs)
 
     def _test():
         test_case.start_time()
@@ -887,7 +880,7 @@ def _run_test(
             name=getattr(func, "__name__", "Test"),
             run_id=test_case.run_id,
             reference_example_id=test_case.example_id,
-            inputs=func_inputs,
+            inputs=test_case.inputs,
             project_name=test_case.test_suite.name,
             exceptions_to_handle=(SkipException,),
             _end_on_exit=False,
@@ -925,10 +918,11 @@ def _run_test(
             "reference_example_id": str(test_case.example_id),
         },
     }
-    with rh.tracing_context(
-        **{**current_context, "metadata": metadata}
-    ), ls_utils.with_optional_cache(
-        cache_path, ignore_hosts=[test_case.test_suite.client.api_url]
+    with (
+        rh.tracing_context(**{**current_context, "metadata": metadata}),
+        ls_utils.with_optional_cache(
+            cache_path, ignore_hosts=[test_case.test_suite.client.api_url]
+        ),
     ):
         _test()
 
@@ -948,8 +942,6 @@ async def _arun_test(
         langtest_extra=langtest_extra,
     )
     _TEST_CASE.set(test_case)
-    func_sig = inspect.signature(func)
-    func_inputs = rh._get_inputs_safe(func_sig, *test_args, **test_kwargs)
 
     async def _test():
         test_case.start_time()
@@ -957,7 +949,7 @@ async def _arun_test(
             name=getattr(func, "__name__", "Test"),
             run_id=test_case.run_id,
             reference_example_id=test_case.example_id,
-            inputs=func_inputs,
+            inputs=test_case.inputs,
             project_name=test_case.test_suite.name,
             exceptions_to_handle=(SkipException,),
             _end_on_exit=False,
@@ -995,10 +987,11 @@ async def _arun_test(
             "reference_example_id": str(test_case.example_id),
         },
     }
-    with rh.tracing_context(
-        **{**current_context, "metadata": metadata}
-    ), ls_utils.with_optional_cache(
-        cache_path, ignore_hosts=[test_case.test_suite.client.api_url]
+    with (
+        rh.tracing_context(**{**current_context, "metadata": metadata}),
+        ls_utils.with_optional_cache(
+            cache_path, ignore_hosts=[test_case.test_suite.client.api_url]
+        ),
     ):
         await _test()
 
@@ -1033,9 +1026,7 @@ def log_inputs(inputs: dict, /) -> None:
                 assert foo(x, y) == 2
     """
     if ls_utils.test_tracking_is_disabled():
-        logger.info(
-            "LANGSMITH_TEST_TRACKING is set to 'false'." " Skipping log_inputs."
-        )
+        logger.info("LANGSMITH_TEST_TRACKING is set to 'false'. Skipping log_inputs.")
         return
     run_tree = rh.get_current_run_tree()
     test_case = _TEST_CASE.get()
@@ -1047,7 +1038,7 @@ def log_inputs(inputs: dict, /) -> None:
         )
         raise ValueError(msg)
     run_tree.add_inputs(inputs)
-    test_case.sync_example(inputs=inputs)
+    test_case.log_inputs(inputs)
 
 
 def log_outputs(outputs: dict, /) -> None:
@@ -1077,9 +1068,7 @@ def log_outputs(outputs: dict, /) -> None:
                 assert result == 2
     """
     if ls_utils.test_tracking_is_disabled():
-        logger.info(
-            "LANGSMITH_TEST_TRACKING is set to 'false'." " Skipping log_outputs."
-        )
+        logger.info("LANGSMITH_TEST_TRACKING is set to 'false'. Skipping log_outputs.")
         return
     run_tree = rh.get_current_run_tree()
     test_case = _TEST_CASE.get()
@@ -1090,11 +1079,12 @@ def log_outputs(outputs: dict, /) -> None:
             "LANGSMITH_TRACING environment variable to 'true')."
         )
         raise ValueError(msg)
+    outputs = _dumpd(outputs)
     run_tree.add_outputs(outputs)
     test_case.log_outputs(outputs)
 
 
-def log_reference_outputs(outputs: dict, /) -> None:
+def log_reference_outputs(reference_outputs: dict, /) -> None:
     """Log example reference outputs from within a pytest test run.
 
     .. warning::
@@ -1122,8 +1112,7 @@ def log_reference_outputs(outputs: dict, /) -> None:
     """
     if ls_utils.test_tracking_is_disabled():
         logger.info(
-            "LANGSMITH_TEST_TRACKING is set to 'false'."
-            " Skipping log_reference_outputs."
+            "LANGSMITH_TEST_TRACKING is set to 'false'. Skipping log_reference_outputs."
         )
         return
     test_case = _TEST_CASE.get()
@@ -1133,7 +1122,7 @@ def log_reference_outputs(outputs: dict, /) -> None:
             "decorated with @pytest.mark.langsmith."
         )
         raise ValueError(msg)
-    test_case.sync_example(outputs=outputs)
+    test_case.log_reference_outputs(reference_outputs)
 
 
 def log_feedback(
@@ -1176,9 +1165,7 @@ def log_feedback(
                 assert result == expected
     """
     if ls_utils.test_tracking_is_disabled():
-        logger.info(
-            "LANGSMITH_TEST_TRACKING is set to 'false'." " Skipping log_feedback."
-        )
+        logger.info("LANGSMITH_TEST_TRACKING is set to 'false'. Skipping log_feedback.")
         return
     if feedback and any((key, score, value)):
         msg = "Must specify one of 'feedback' and ('key', 'score', 'value'), not both."
@@ -1282,9 +1269,7 @@ def trace_feedback(
                 assert "hello" in response.choices[0].message.content.lower()
     """  # noqa: E501
     if ls_utils.test_tracking_is_disabled():
-        logger.info(
-            "LANGSMITH_TEST_TRACKING is set to 'false'." " Skipping log_feedback."
-        )
+        logger.info("LANGSMITH_TEST_TRACKING is set to 'false'. Skipping log_feedback.")
         yield None
         return
     parent_run = rh.get_current_run_tree()
@@ -1316,3 +1301,25 @@ def _stringify(x: Any) -> str:
         return dumps_json(x).decode("utf-8", errors="surrogateescape")
     except Exception:
         return str(x)
+
+
+def _dumpd(x: Any) -> Any:
+    """Serialize LangChain Serializable objects."""
+    dumpd = _get_langchain_dumpd()
+    if not dumpd:
+        return x
+    try:
+        serialized = dumpd(x)
+        return serialized
+    except Exception:
+        return x
+
+
+@functools.lru_cache
+def _get_langchain_dumpd() -> Optional[Callable]:
+    try:
+        from langchain_core.load import dumpd
+
+        return dumpd
+    except ImportError:
+        return None
